'''
Perform magnetoentropic mapping of magnetic materials based on DC
magnetization data.
'''

from _meta import __author__, __copyright__, __license__, __version__
from .magentro import MagentroData
from . import errors
from . import typedefs

__all__ = ['MagentroData', 'errors', 'typedefs']



'''Contains convenience functions for working with Pint.'''

from __future__ import annotations

from typing import Optional, Sequence

import numpy as np
from numpy.typing import NDArray
import pandas as pd
from pint import UnitRegistry, Quantity

from .typedefs import ColumnDataDict, DataProp
from ._classvars import RAW_DF_COLUMNS, RAW_DF_DEFAULT_UNITS

def pmag(pint_series: pd.Series) -> NDArray[np.float64]:
    '''Return the underlying float ``ndarray`` of a |Series| containing a ``PintArray``.'''
    return pint_series.values.quantity.magnitude

def punits(pint_series: pd.Series) -> str:
    '''Return the short, pretty-printed units ``str`` of a |Series| containing a ``PintArray``.'''
    return f'{pint_series.values.quantity.units:~P}'

def df_from_units(
    col_names: Sequence[str],
    units: Sequence[str],
    data: Optional[NDArray[np.float64]] = None
    ) -> pd.DataFrame:
    '''Construct a |DataFrame| given column names and units, optionally with data.'''

    if data is not None:
        return pd.DataFrame(
            data={
                col: pd.Series(data[:, i], dtype=f'pint[{unit}]')
                for col, unit, i in zip(col_names, units, range(len(col_names)))
            }
        )

    return pd.DataFrame(
        data={col: pd.Series([], dtype=f'pint[{unit}]') for col, unit in zip(col_names, units)}
    )

def create_per_mass_cols(
    df: pd.DataFrame,
    cols: dict[str, str],
    units: dict[str, str],
    sample_mass: Quantity
    ) -> None:
    '''Create per-mass columns in `df`.'''

    convert_series(
        df.loc[:, cols['M']] / sample_mass,
        df,
        cols['M_per_mass'],
        units['M_per_mass']
    )

    convert_series(
        df.loc[:, cols['M_err']] / sample_mass,
        df,
        cols['M_per_mass_err'],
        units['M_per_mass_err']
    )

def convert_df(
    initial_df: pd.DataFrame,
    final_df: pd.DataFrame,
    i_names: Sequence[str],
    f_names: Sequence[str],
    f_unit_strs: Sequence[str]
    ) -> None:
    '''
    Fill columns in `final_df` (`f_names`) with data converted from
    `initial_df` columns (`i_names`).

    Conversion is from units contained in `initial_df`
    to units given by `f_unit_strs`.
    '''

    for i, f, units in zip(i_names, f_names, f_unit_strs):
        convert_series(initial_df.loc[:, i], final_df, f, units)

def convert_series(a_series: pd.Series, b_df: pd.DataFrame, b_col: str, units: str) -> None:
    '''Fill `b_col` in `b_df` with `a_series` converted to `units`.'''
    b_df.loc[:, b_col] = a_series.pint.to(units)

def convert_processed_raw(
    processed_df: pd.DataFrame,
    cols_to_convert: Sequence[str],
    raw_units: Sequence[str]
    ) -> None:
    '''Convert `cols_to_convert` to `raw_units` in `processed_df` in-place.'''

    convert_df(
        processed_df, processed_df,
        cols_to_convert, cols_to_convert,
        raw_units
    )

def get_raw_conv_t_h(df) -> NDArray[np.float64]:
    '''Convert T and H to default raw data units and return values as 2D ``ndarray``.'''

    r_cols, r_units = RAW_DF_COLUMNS, RAW_DF_DEFAULT_UNITS

    t_h_cols = [r_cols['T'], r_cols['H']]
    t_h_units = [r_units['T'], r_units['H']]

    t_h_raw_converted_df = df_from_units(t_h_cols, t_h_units)
    convert_df(df, t_h_raw_converted_df, t_h_cols, t_h_cols, t_h_units)

    return t_h_raw_converted_df.pint.dequantify().values

def get_latex_units(
    ureg: UnitRegistry,
    data_prop: DataProp,
    units: ColumnDataDict) -> tuple[str, str, str]:
    '''Return LaTeX strings for `units`.'''

    T_units = f'{ureg(units["T"]).units:~L}'
    H_units = f'{ureg(units["H"]).units:~L}'
    prop_units = f'{ureg(units[data_prop]).units:~L}'

    return T_units, H_units, prop_units



'''Contains functions to group data by field or temperature.'''

from __future__ import annotations

from typing import Union, Optional

import numpy as np
from numpy.typing import NDArray
import pandas as pd
from pandas.core.groupby.generic import DataFrameGroupBy
from numba import njit

from magentropy.errors import MissingDataError

from .typedefs import Presets
from ._classvars import RAW_DF_COLUMNS, DEFAULT_PRESETS

def test_grouping(
    raw_df: pd.DataFrame,
    presets: Presets
    ) -> tuple[dict[str, Union[NDArray[np.float64], np.int64, np.float64]], DataFrameGroupBy]:
    '''For use in ``magentro.test_grouping`` and ``magentro.test_grouping_``.'''

    fds = presets['fields']
    dms = presets['decimals']
    md = presets['max_diff']

    return (
        {'fields': fds, 'decimals': dms, 'max_diff': md},
        group_by(
            raw_df,
            group_col=RAW_DF_COLUMNS['H'],
            groups=fds,
            decimals=dms,
            max_diff=md
        )
    )

def group_by(
    df: pd.DataFrame,
    group_col: str,
    groups: NDArray[np.float64] = DEFAULT_PRESETS['fields'],
    decimals: np.int64 = DEFAULT_PRESETS['decimals'],
    max_diff: np.float64 = DEFAULT_PRESETS['max_diff'],
    match_col: Optional[str] = None,
    match_values: Optional[NDArray[np.float64]] = None
    ) -> DataFrameGroupBy:
    '''
    Group data by temperature or field strength.

    Parameters
    ----------
    df : DataFrame
        |DataFrame| to group. Columns should be ordinary |Series|s, without
        ``PintArray``\s, e.g. from ``self.raw_data`` instead of
        ``self._raw_data``.
    group_col : str
        Name of column to use for grouping.
    groups : ndarray
        Expected groups. If empty, the groups are determined automatically
        based on `max_diff` and/or `decimals`.
    decimals : int
        The decimal place to which to round the automatically determined groups.
        (A negative integer specifies the number of positions to the left of the
        decimal point. See :func:`numpy.around` `decimal` parameter documentation.)
        Ignored if `groups` is not empty. If groups is empty and `max_diff` is
        ``np.inf``, groups will be determined by rounding to this decimal place.
    max_diff : float
        If `groups` is not empty, `max_diff` is the maximum difference allowed
        between each value in the `group_col` column and the closest group in
        `groups`.
        Values too far away from any group will be omitted, unless max_diff is
        ``np.inf``.
        If `groups` is empty, `max_diff` is the maximum difference allowed
        between any two items in each group, which is used to determine groups
        automatically.
        Generally, `decimals` is enough to determine groups, but `max_diff` can
        be used for finer control.
    match_col : str, optional
        Name of column to use for matching values within group.
    match_values : ndarray, optional
        Values to match within each group.
        Only the ``len(match_values)`` observations in `match_col` of each group
        that are closest to one of `match_values` will be kept.
        Other observations will be indicated as omitted by filling with ``NaN``.

    Returns
    -------
    grouped_by : DataFrameGroupBy
    '''

    if len(groups) != 0:
        # take the (unique) groups supplied
        unique_groups = np.unique(groups)

        if max_diff != np.inf:
            # remove values that are too far away from any group
            all_diffs = abs(unique_groups[:, np.newaxis] - df.loc[:, group_col].values)
            faraway_mask = np.min(all_diffs, axis=0) <= max_diff
            df = df.loc[faraway_mask, :]

    elif max_diff != np.inf:
        # automatically determine groups
        unique_groups = find_unique_groups(df.loc[:,group_col].values, max_diff, decimals)
    else:
        # determine groups solely by rounding
        unique_groups = np.unique(
            df.loc[:, group_col].round(decimals=decimals)
        )

    # perform the groupby based on which values are closest to a group
    grouped_by = df.copy().groupby(
        lambda x: unique_groups[abs(unique_groups - df.loc[:, group_col].loc[x]).argmin()],
        group_keys=False
    )

    if match_col is not None and match_values is not None:
        match_col = str(match_col)
        match_values = np.unique(match_values)
        grps = list(grouped_by.groups.keys())
        df = grouped_by.apply(
            match_closest_values,
            group_col,
            match_col,
            match_values,
            grps
        )
        grouped_by = df.groupby(
            lambda x: unique_groups[abs(unique_groups - df.loc[:, group_col].loc[x]).argmin()],
            group_keys=False
        )

    return grouped_by

@njit(cache=True)
def find_unique_groups(
    temps_or_fields: NDArray[np.float64],
    max_diff: np.float64,
    decimals: np.int64
    ) -> NDArray[np.float64]:
    '''Return unique temperatures or fields based on `max_diff`.'''

    # order values to group
    temps_or_fields = np.sort(temps_or_fields)

    if len(temps_or_fields) < 2:
        return temps_or_fields

    # groups will be a list of lists (of varying length)
    groups = [[temps_or_fields[0]]]
    group_idx = 0

    for tf in temps_or_fields[1:]:
        # create new group if difference between value and first item of current group is large
        if abs(tf - groups[group_idx][0]) > max_diff:
            groups.append([np.float64(x) for x in range(0)]) # for numba's benefit
            group_idx += 1

        # append value to current group
        groups[group_idx].append(tf)

    # take the mean of each sub-list, then round
    rounded_mean_groups = np.zeros(len(groups), dtype=np.float64)
    np.around(
        np.array([np.array(g).mean() for g in groups], dtype=np.float64),
        decimals,
        rounded_mean_groups
    )

    return rounded_mean_groups

def match_closest_values(
    da: pd.DataFrame,
    group_col: str,
    match_col: str,
    match_values: NDArray[np.float64],
    grps: list[np.float64]
    ) -> pd.DataFrame:
    '''
    Passed to ``DataFrameGroupBy.apply``. Primarily for grouping by temperature.

    Finds the closest observations in `match_col` of `da` to any one of
    `match_values`, then limits data to one observation per match value by
    taking the one with the closest value in `group_col` to `grps[0]` for
    each match value.
    '''

    # get the current group and remove from the list
    try:
        grp = grps.pop(0)
    except IndexError as e:
        raise MissingDataError(
            'Not all groups found a match. '
            'Check the supplied field or temperature groups to ensure they are '
            'appropriate for the data.'
        ) from e

    # array of lowest distance from a value in match_values for each observation
    match_diffs = abs(match_values[:, np.newaxis] - da.loc[:, match_col].values)
    min_idx = np.argmin(match_diffs, axis=0)

    group_values = da.loc[:, group_col].values
    group_diffs = abs(group_values - grp)
    min_distances = np.sqrt(np.min(match_diffs, axis=0)**2 + np.min(group_diffs)**2)
    closest_idxs = np.zeros(len(match_values), dtype=np.int64)

    for i in range(len(match_values)):
        group_values_subset = group_values.copy()
        group_values_subset[min_idx != i] = np.nan
        group_distances_subset = min_distances.copy()
        group_distances_subset[min_idx != i] = np.nan
        closest_idxs[i] = np.nanargmin(group_distances_subset)

    da = da.iloc[closest_idxs, :]

    return da



'''Exception classes.'''

from __future__ import annotations

__all__ = ['MagentroError', 'UnitError', 'MissingDataError']

class MagentroError(Exception):
    '''Base exception class for |MagentroData|.'''

class UnitError(ValueError, MagentroError):
    '''Exception class for invalid units or conversions.'''

class MissingDataError(MagentroError):
    '''Exception class for attempting to plot or operate on empty data.'''



'''
Contains functions used for simulation, smoothing, interpolation, differentiation, and integration.
'''

from __future__ import annotations

from typing import Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
import pandas as pd
from pint import Quantity
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import griddata, RegularGridInterpolator
from numba import njit

from .typedefs import Kwargs, MatchErrStr
from ._classvars import RAW_DF_COLUMNS, RAW_DF_DEFAULT_UNITS, PROCESSED_DF_COLUMNS
from . import _pint_utils as pu
from . import _smoothing

@njit(cache=True)
def sim_func(
    t: NDArray[np.float64],
    h: NDArray[np.float64],
    m_max: np.float64,
    slope: np.float64,
    bump_height: np.float64
    ) -> NDArray[np.float64]:
    '''
    Compute simulated data.

    Model function is a decreasing logistic function with a tiny Gaussian bump.

    Decreasing logistic function: L / (1 + e^[-k(t-t0)])

    Gaussian: A e^(-(t-mu)^2 / a^2)
    '''

    # choose minimum moment to be 1/len(h) of m_max
    m_maxes = np.linspace(m_max / len(h), m_max, len(h))

    # Choose x0 to be two-thirds of the temperature range
    t_range = t.max() - t.min()
    t0 = t.min() + (2./3.) * t_range

    # derivative at t0 is log_max * k / 4, so scale k inversely with t_range
    k = -4. / t_range * slope

    gaussian_amp = m_max * bump_height

    # choose the Gaussian spread to be about one-fifth of the temperature range
    gaussian_spr = t_range / 5.

    # make the Gaussian center vary with field
    mus = t.min() + np.linspace(t_range * (2./3.), t_range / 3., len(h))

    m_grid = np.zeros((len(h), len(t)), dtype=np.float64)

    for i in range(len(h)):
        m_grid[i, :] = (
            m_maxes[i] / (1. + np.exp(-k * (t-t0)))
            + gaussian_amp * np.exp(-((t-mus[i]) / gaussian_spr)**2)
        )

    return m_grid

def smooth_diff_int_processed(
    used_fields: NDArray[np.float64],
    df_arrays: tuple[NDArray[np.float64], ...],
    n: np.int64,
    tr: NDArray[np.float64],
    d: np.int64,
    lms: NDArray[np.float64],
    lm_g: np.float64,
    we: bool,
    me: Union[bool, ArrayLike, MatchErrStr],
    mk: Kwargs,
    az: bool,
    current_p_units: dict[str, str],
    sample_mass: Quantity
    ) -> pd.DataFrame:
    '''Perform the smoothing, differentiation, and integration on the processed data.'''

    r_units = RAW_DF_DEFAULT_UNITS
    p_cols = PROCESSED_DF_COLUMNS

    # smooth in the supplied units for T and M, respectively
    smooth_output = create_output_arr(used_fields, n, np.linspace(tr[0], tr[1], n))
    _smoothing.smooth_processed(used_fields, df_arrays, smooth_output, d, lms, lm_g, we, me, mk)

    # fill processed df with the smoothed data in the supplied units
    smooth_output = smooth_output.reshape((len(used_fields)+1) * n, 8)
    processed_df = pu.df_from_units(p_cols.values(), current_p_units.values(), smooth_output)

    # convert T, H, M, and M_err to the default raw data units
    pu.convert_processed_raw(
        processed_df,
        (p_cols['T'], p_cols['H'], p_cols['M'], p_cols['M_err']),
        (r_units['T'], r_units['H'], r_units['M'], r_units['M_err'])
    )

    current_p_units['T'] = r_units['T']
    current_p_units['H'] = r_units['H']
    current_p_units['M'] = r_units['M']
    current_p_units['M_err'] = r_units['M_err']

    # compute per-mass columns
    pu.create_per_mass_cols(processed_df, p_cols, r_units, sample_mass)

    # perform the differentiation and integration in the default raw data units
    diff_int_output = processed_df.pint.dequantify().values.reshape(len(used_fields)+1, n, 8)
    diff_int_processed(diff_int_output)
    diff_int_output = diff_int_output.reshape((len(used_fields)+1) * n, 8)

    # remove zeros at beginning if add_zeros is False
    if not az:
        diff_int_output = diff_int_output[n:, :]

    processed_df = pu.df_from_units(p_cols.values(), current_p_units.values(), diff_int_output)

    # convert dM_dT and Delta_SM, which are in units directly constructed from the
    # defaults for T, H, and M, to the default (standard cgs) raw data units.
    pu.convert_processed_raw(
        processed_df,
        (p_cols['dM_dT'], p_cols['Delta_SM']),
        (r_units['dM_dT'], r_units['Delta_SM'])
    )

    return processed_df

@njit(parallel=True, cache=True)
def create_output_arr(
    used_fields: NDArray[np.float64],
    n: np.int64,
    t_grid: NDArray[np.float64]
    ) -> NDArray[np.float64]:
    '''
    Return an output array to hold the processed data.

    Like a stack of df values: (0 + fields) x npoints x columns

    The prepended field row holds data for zero field, where moment is zero
    (for integration).

    The first depth layer, output[:, :, 0], holds the temperature grid for
    each field.

    The second depth layer excluding zero field, output[1:, :, 1], holds the
    field values.

    M_err and M_per_mass_err are NaN.
    '''

    nfields = len(used_fields)

    output = np.zeros((nfields + 1, n, 8))
    output[:, :, 0] = t_grid.repeat(nfields + 1).reshape((n, nfields + 1)).T
    output[1:, :, 1] = used_fields.repeat(n).reshape((nfields, n))
    output[:, :, 3] = np.nan
    output[:, :, 5] = np.nan

    return output

def diff_int_processed(output: NDArray[np.float64]) -> None:
    '''
    Perform the differentiation and integration for the processed data.

    Data indices corresponding to columns are
    T: 0, H: 1, M_per_mass: 4, dM_dT: 6, Delta_SM: 7

    output[0,:,0] is a 1D array of the temperature grid

    output[:,0,1] is a 1D array of the field values
    '''

    # gradient along the npoints temperatures (axis=1)
    output[1:, :, 6] = np.gradient(output[1:,:,4], output[0,:,0], axis=1, edge_order=2)

    # integrate derivatives along the fields (axis=0)
    output[:, :, 7] = cumulative_trapezoid(output[:,:,6], x=output[:,0,1], axis=0, initial=0.0)

def interp_diff_int_raw(
    raw_df: pd.DataFrame,
    n: np.int64,
    temp_diff: np.float64,
    fields: NDArray[np.float64],
    t_h_raw_conv_values: NDArray[np.float64]
    ) -> None:
    '''Perform the interpolation, differentiation, and integration on the raw data.'''

    t_lin, h_lin, m_per_mass_grid = interp_setup(raw_df, n, temp_diff, fields, t_h_raw_conv_values)

    dm_dt_call, delta_sm_call = interp_callables(t_lin, h_lin, m_per_mass_grid)

    raw_df[RAW_DF_COLUMNS['dM_dT']] = dm_dt_call(t_h_raw_conv_values)
    raw_df[RAW_DF_COLUMNS['Delta_SM']] = delta_sm_call(t_h_raw_conv_values)

def interp_setup(
    raw_df: pd.DataFrame,
    n: np.int64,
    temp_diff: np.float64,
    fields: NDArray[np.float64],
    t_h_raw_conv_values: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    '''Return T and H linspaces, and M_per_mass grid.'''

    t_min, t_max = t_h_raw_conv_values[:, 0].min(), t_h_raw_conv_values[:, 0].max()
    num_temps = np.int64(np.around(n * (t_max - t_min) / temp_diff))

    t_lin = np.linspace(t_min, t_max, num_temps)
    h_lin = np.concatenate(([0.0], fields, [t_h_raw_conv_values[:, 1].max()]))

    t_grid, h_grid = np.meshgrid(t_lin, h_lin, indexing='ij')

    # interpolated moment grid
    m_per_mass_grid = griddata(
        (t_h_raw_conv_values[:, 0], t_h_raw_conv_values[:, 1]),
        raw_df[RAW_DF_COLUMNS['M_per_mass']],
        (t_grid, h_grid),
        method='linear'
    )

    # same, but with nearest method instead of linear
    m_per_mass_nearest = griddata(
        (t_h_raw_conv_values[:, 0], t_h_raw_conv_values[:, 1]),
        raw_df[RAW_DF_COLUMNS['M_per_mass']],
        (t_grid, h_grid),
        method='nearest'
    )

    # fill any missing values in the "linear" grid with those from the "nearest" grid
    m_per_mass_nan_idx = np.isnan(m_per_mass_grid)
    m_per_mass_grid[m_per_mass_nan_idx] = m_per_mass_nearest[m_per_mass_nan_idx]

    return t_lin, h_lin, m_per_mass_grid

def interp_callables(
    t_lin: NDArray[np.float64],
    h_lin: NDArray[np.float64],
    m_per_mass_grid: NDArray[np.float64]
    ) -> tuple[RegularGridInterpolator, RegularGridInterpolator]:
    '''
    Compute dM_dT and Delta_SM grids and return callable
    ``RegularGridInterpolator``\s for interpolating dM_dT and Delta_SM,
    respectively, back to original T and H points.
    '''

    # compute dM_dT and Delta_SM from the grid
    dm_dt_grid = np.gradient(m_per_mass_grid, t_lin, axis=0, edge_order=2)
    delta_sm_grid = cumulative_trapezoid(dm_dt_grid, x = h_lin, axis=1, initial=0.0)

    # interpolate dM_dT back to the original T, H points
    dm_dt_call = RegularGridInterpolator(
        (t_lin, h_lin),
        dm_dt_grid,
        method='nearest',
        bounds_error=False
    )

    # interpolate Delta_SM back to the original T, H points
    delta_sm_call = RegularGridInterpolator(
        (t_lin, h_lin),
        delta_sm_grid,
        method='nearest',
        bounds_error=False
    )

    return dm_dt_call, delta_sm_call

def bootstrap(
    used_fields: NDArray[np.float64],
    df_arrays: tuple[NDArray[np.float64], ...],
    n: np.int64,
    tr: NDArray[np.float64],
    d: np.int64,
    lms: NDArray[np.float64],
    we: bool,
    az: bool,
    current_p_units: dict[str, str],
    sample_mass: Quantity,
    n_bootstrap: np.int64,
    rng: np.random.Generator
    ) -> pd.DataFrame:
    '''Perform the bootstrap calculation on the processed data.'''

    r_units = RAW_DF_DEFAULT_UNITS
    p_cols = PROCESSED_DF_COLUMNS

    # bootstrap in the supplied units for T and M, respectively
    bootstrap_output = create_output_arr(used_fields, n, np.linspace(tr[0], tr[1], n))
    _smoothing.bootstrap(
        used_fields, df_arrays, bootstrap_output,
        d, lms, we, n_bootstrap, rng
    )

    # fill processed df with the smoothed data in the supplied units
    bootstrap_output = bootstrap_output.reshape((len(used_fields)+1) * n, 8)

    bootstrap_df = pu.df_from_units(p_cols.values(), current_p_units.values(), bootstrap_output)

    # convert T, H, M, and M_err to the default raw data units
    pu.convert_processed_raw(
        bootstrap_df,
        (p_cols['T'], p_cols['H'], p_cols['M'], p_cols['M_err']),
        (r_units['T'], r_units['H'], r_units['M'], r_units['M_err'])
    )

    current_p_units['T'] = r_units['T']
    current_p_units['H'] = r_units['H']
    current_p_units['M'] = r_units['M']
    current_p_units['M_err'] = r_units['M_err']

    # compute per-mass columns
    pu.create_per_mass_cols(bootstrap_df, p_cols, r_units, sample_mass)

    bootstrap_output = bootstrap_df.pint.dequantify().values

    # remove zeros at beginning if add_zeros is False
    if not az:
        bootstrap_output = bootstrap_output[n:, :]

    bootstrap_df = pu.df_from_units(p_cols.values(), current_p_units.values(), bootstrap_output)

    return bootstrap_df



'''MagentroPy package metadata.'''

# keep consistent with pyproject.toml
__project__ = 'MagentroPy'
__author__ = 'Soren Bear'
__email__ = 'sorenbear@berkeley.edu'
__copyright__ = f'2022-2023, {__author__}'
__license__ = 'MIT'
__version__= '0.1.6'
__repo_url__ = 'https://github.com/sorenbear/magentropy'
